library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.std_logic_unsigned.ALL;

entity configurable_adder_sub is
	generic ( 
		OPERAND_LENGTH : integer	:= 32;
		EXPONENT_LENGTH : integer	:= 8;
		MANTISSA_LENGTH : integer	:= 23
		);
	Port ( a 		: in  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			b 			: in  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			clk		:	in std_logic;
			reset		:	in std_logic;
			new_data	:	in std_logic;
			operation	: in std_logic_vector(2 downto 0);
			result 	: out  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			rdy		: out	std_logic);
end configurable_adder_sub;

architecture Behavioral of configurable_adder_sub is

type state_type is (s0,s1,s2,s3,s4,s5,s6,s7,s8);  --type of state machine.
signal current_s,next_s: state_type;  --current and next state declaration.

-- Signals for adder
signal res_exponent				:	std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');
signal res_mantissa 				:	std_logic_vector(MANTISSA_LENGTH+1 downto 0) := (others => '0');
signal res_sign					:	std_logic	:= '0';
signal diff_exponent				:	std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');

signal exponent_1					:	std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');
signal exponent_2					:	std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');
signal mantissa_1					:	std_logic_vector(MANTISSA_LENGTH-1 downto 0) := (others => '0');
signal mantissa_2					:	std_logic_vector(MANTISSA_LENGTH-1 downto 0) := (others => '0');

signal local_op					:	std_logic_vector(2 downto 0) := (others => '0');
signal normalization				:	std_logic := '0';
signal de_normalized_1			:	std_logic := '0';
signal de_normalized_2			:	std_logic := '0';



begin

process (clk,reset)
begin
	if (reset='1') then
		current_s <= s0;  --default state on reset.
	elsif (rising_edge(clk)) then
		current_s <= next_s;   --state change.
	end if;
end process;

process (current_s, a, b, new_data, operation)
begin
	case current_s is
		when s0 =>
			res_exponent <= (others => '0');
			res_mantissa <= (others => '0');
			res_sign <= '0';
			diff_exponent <= (others => '0');
			exponent_1 <= (others => '0');
			exponent_2 <= (others => '0');
			mantissa_1 <= (others => '0');
			mantissa_2 <= (others => '0');
			rdy <= '0';
			de_normalized_1 <= '0';
			de_normalized_2 <= '0';

		
			if new_data = '1' then 
				next_s <= s1;
			else next_s <= s0;
			end if;
		when s1 =>
		
			if(	(a(OPERAND_LENGTH-1) = '0' and b(OPERAND_LENGTH-1) = '0' and operation = "000") 
				or (a(OPERAND_LENGTH-1) = '0' and b(OPERAND_LENGTH-1) = '1' and operation = "001")) then
				
				exponent_1 	<= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				exponent_2 	<= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				mantissa_1 	<= a(MANTISSA_LENGTH-1 downto 0);
				mantissa_2 	<= b(MANTISSA_LENGTH-1 downto 0);
				local_op		<= "000";
				res_sign		<= '0';
				next_s <= s3;
				
			elsif((a(OPERAND_LENGTH-1) = '0' and b(OPERAND_LENGTH-1) = '0' and operation = "001") 
				or (a(OPERAND_LENGTH-1) = '0' and b(OPERAND_LENGTH-1) = '1' and operation = "000")) then
				
				exponent_1 	<= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				exponent_2 	<= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				mantissa_1 	<= a(MANTISSA_LENGTH-1 downto 0);
				mantissa_2 	<= b(MANTISSA_LENGTH-1 downto 0);
				local_op		<= "001";
				res_sign		<= '0';
				next_s <= s2;
				
			elsif((a(OPERAND_LENGTH-1) = '1' and b(OPERAND_LENGTH-1) = '0' and operation = "000") 
				or (a(OPERAND_LENGTH-1) = '1' and b(OPERAND_LENGTH-1) = '1' and operation = "001")) then
				
				exponent_1 	<= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				exponent_2 	<= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				mantissa_1 	<= b(MANTISSA_LENGTH-1 downto 0);
				mantissa_2 	<= a(MANTISSA_LENGTH-1 downto 0);
				local_op		<= "001";
				res_sign <= '1';
				next_s <= s2;
				
			else
			
				exponent_1 	<= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				exponent_2 	<= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				mantissa_1 	<= a(MANTISSA_LENGTH-1 downto 0);
				mantissa_2 	<= b(MANTISSA_LENGTH-1 downto 0);
				local_op		<= "000";
				res_sign		<= '1';
				next_s <= s3;
				
			end if;
		
		-- Diff with sign bit
		when s2 =>
				if((exponent_1&mantissa_1) > (exponent_2&mantissa_2)) then
					res_sign <= '0';
					diff_exponent <= exponent_1 - exponent_2;
					next_s <= s4;
				else
					res_sign <= '1';
					diff_exponent <= exponent_2 - exponent_1;
					next_s <= s5;
				end if;
		-- Diff without sign bit
		when s3 =>
			if((exponent_1&mantissa_1) > (exponent_2&mantissa_2)) then
				diff_exponent <= exponent_1 - exponent_2;
				next_s <= s4;
			else
				diff_exponent <= exponent_2 - exponent_1;
				next_s <= s5;
			end if;
			
			
		when s4 =>
			-- Shift the smallest mantissa x places to the right if the difference between the exponents are bigger then 0
			-- If the mantissa is shifted, it is denormalized, going from 1.xxxx to 0.xxxx
			if(diff_exponent > 0 and diff_exponent <= MANTISSA_LENGTH-1) then
				mantissa_2 <= std_logic_vector(unsigned(mantissa_2) srl to_integer(unsigned(diff_exponent)));
				mantissa_2(MANTISSA_LENGTH-to_integer(unsigned(diff_exponent))) <= '1';
				de_normalized_2 <= '1';
				res_exponent <= exponent_1;
				next_s <= s6;
			elsif(diff_exponent > (MANTISSA_LENGTH-1)) then
				-- a >> b
				res_exponent <= exponent_1;
				res_mantissa(MANTISSA_LENGTH-1 downto 0) <= mantissa_1;
				next_s <= s8;
			else
				next_s <= s6;
			end if;
			
			
			
		when s5 =>
			-- Shift the smallest mantissa x places to the right if the difference between the exponents are bigger then 0
			-- If the mantissa is shifted, it is denormalized, going from 1.xxxx to 0.xxxx
			if(diff_exponent > 0 and diff_exponent <= MANTISSA_LENGTH-1) then
				mantissa_1 <= std_logic_vector(unsigned(mantissa_1) srl to_integer(unsigned(diff_exponent)));
				mantissa_1(MANTISSA_LENGTH-to_integer(unsigned(diff_exponent))) <= '1';
				de_normalized_1 <= '1';
				res_exponent <= exponent_2;
				next_s <= s6;
			elsif(diff_exponent > (MANTISSA_LENGTH-1)) then
				-- a >> b
				res_exponent <= exponent_2;
				res_mantissa(MANTISSA_LENGTH-1 downto 0) <= mantissa_2;
				next_s <= s8;
			else
				next_s <= s6;
				res_exponent <= exponent_2;
			end if;
			
			
		when s6 =>
			-- Check for addition or subtraction
			if (local_op(0) = '0') then
				if(de_normalized_1 = '1') then
					res_mantissa <= ("00" & mantissa_1) + ("01" & mantissa_2);
				elsif(de_normalized_2 = '1') then
					res_mantissa <= ("01" & mantissa_1) + ("00" & mantissa_2);
				else
					res_mantissa <= ("01" & mantissa_1) + ("01" & mantissa_2);
				end if;
				
			elsif (local_op(0) = '1') then
				if(de_normalized_1 = '1') then
					res_mantissa <= ("01" & mantissa_2)-("00" & mantissa_1);--("00" & mantissa_1) - ("01" & mantissa_2);
				elsif(de_normalized_2 = '1') then
					res_mantissa <= ("01" & mantissa_1) - ("00" & mantissa_2);
				else
					res_mantissa(MANTISSA_LENGTH-1 downto 0) <= mantissa_1(MANTISSA_LENGTH-1 downto 0) - mantissa_2(MANTISSA_LENGTH-1 downto 0);
				end if;
			else
			end if;
			
			next_s <= s7;
			
		when s7 =>
		
			-- Test if the resulting mantissa is normalized
			if(res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) > 1) then
				-- Needs to be normalized, mantissa bigger then 2
				normalization <= '0';
				
				-- Exponent has to be added
				res_exponent <= res_exponent + res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) - 1;
				-- Resulting mantissa is shifted 1 to right, normalized
				res_mantissa(MANTISSA_LENGTH-1 downto 0) <= res_mantissa(MANTISSA_LENGTH downto 1);
			
			elsif (res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) = 1) then
				-- Mantissa is normalized
				normalization <= '0';
				--res_mantissa <= res_mantissa(MANTISSA_LENGTH-1 downto 0);
				
			else 
				normalization <= '0';
				
				-- Find the position of the leading one
				-- Is this the best way to do it????
				for i in MANTISSA_LENGTH-1 downto 0 loop
					if(res_mantissa(i) = '1') then
					-- Leading one found
						if((MANTISSA_LENGTH-i) > res_exponent) then
							res_mantissa <= (others => '0');
						else
							res_mantissa <= std_logic_vector(unsigned(res_mantissa) sll (MANTISSA_LENGTH-i));
							res_exponent <= res_exponent - (MANTISSA_LENGTH-i);
						end if;
						
--						a_adder(EXPONENT_LENGTH-1 downto 0) <= big_exponent;
--						b_adder(4 downto 0) <= std_logic_vector(to_unsigned(MANTISSA_LENGTH-i,5));
--						add_adder <= '0';
--						ce_adder <= '1';
						exit;
					else
						
					end if;
				end loop;				
				
			end if;
			next_s <= s8;
			
		when s8 =>
			result <= res_sign & res_exponent & res_mantissa(MANTISSA_LENGTH-1 downto 0);
			rdy <= '1';
			next_s <= s0;
		when others =>
			
			next_s <= s0;
	end case;
end process;


end Behavioral;

