
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.std_logic_unsigned.ALL;

entity configurable_adder_unsigned is
	generic ( 
		OPERAND_LENGTH : integer	:= 15;
		EXPONENT_LENGTH : integer	:= 5;
		MANTISSA_LENGTH : integer	:= 9
		);
	Port ( a 		: in  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			b 			: in  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			clk		:	in std_logic;
			reset		:	in std_logic;
			new_data	:	in std_logic;
			operation	: in std_logic_vector(2 downto 0);
			result 	: out  STD_LOGIC_VECTOR (OPERAND_LENGTH-1 downto 0);
			rdy		: out	std_logic);
end configurable_adder_unsigned;

architecture Behavioral of configurable_adder_unsigned is

type state_type is (s0,s1,s2,s3,s4,s5,s6,s7,s8);  --type of state machine.
signal current_s,next_s: state_type;  --current and next state declaration.

-- Signals for adder
signal result_exponent	:	std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');

signal res_mantissa : std_logic_vector(MANTISSA_LENGTH+1 downto 0) := (others => '0');
signal big_exponent : std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');
signal small_exponent : std_logic_vector(EXPONENT_LENGTH-1 downto 0) := (others => '0');
signal big_mantissa : std_logic_vector(MANTISSA_LENGTH-1 downto 0) := (others => '0');
signal small_mantissa : std_logic_vector(MANTISSA_LENGTH-1 downto 0) := (others => '0');
signal normalization		: std_logic := '0';
signal sign					: std_logic	:= '0';
signal sub_neg				: std_logic := '0';

signal clk_node			: std_logic := '0';
signal delay_clk			: std_logic := '0';
signal normalization_factor			: std_logic_vector(1 downto 0);
signal sig_operation		: std_logic := '0';

begin

process (clk,reset)
begin
	if (reset='1') then
		current_s <= s0;  --default state on reset.
	elsif (rising_edge(clk)) then
		current_s <= next_s;   --state change.
	end if;
end process;

process (current_s, a, b, new_data, operation)
begin
	case current_s is
		when s0 =>
			rdy <= '0';
			res_mantissa <= (others => '0');
			big_exponent <= (others => '0');
			small_exponent <= (others => '0');
			big_mantissa <= (others => '0');
			small_mantissa <= (others => '0');
			result <= (others => '0');
			normalization <= '0';
			sign	<= '0';
			sub_neg <= '0';
			
			if new_data = '1' then 
				next_s <= s1;
			else next_s <= s0;
			end if;
		when s1 =>
			result <= (others => '0');
			rdy <= '0';
			res_mantissa <= (others => '0');
			sign <= '0';
			
			
			if(operation(0) = '0') then
				-- order do not matter
				next_s <= s2;
			elsif(operation(0) = '1') then
				-- order matter
			end if;
			
			-- Allocate the biggest number to the biggest exponent and mantissa visa verca for further
			-- operations
			-- B bigger then A
			if (b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH) > a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH)) then 
				big_exponent <= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				small_exponent <= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				big_mantissa <= b(MANTISSA_LENGTH-1 downto 0);
				small_mantissa <= a(MANTISSA_LENGTH-1 downto 0);
				
				if(operation(0) = '0') then
					-- order do not matter
					next_s <= s2;
				elsif(operation(0) = '1') then
					-- order matter
					-- If minus and minus
					if(b(OPERAND_LENGTH-1) = '1') then
						sign <= '0';
						sub_neg <= '1';
					else
						sign <= '1';
					end if;
				else
				end if;
			-- A bigger or equal to B
			else	
			
			
				-- A is negative
				if(a(OPERAND_LENGTH-1) = '1') then
					sign <= '1';
				end if;
				big_exponent <= a(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				small_exponent <= b(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH);
				big_mantissa <= a(MANTISSA_LENGTH-1 downto 0);
				small_mantissa <= b(MANTISSA_LENGTH-1 downto 0);
			end if;
			
			next_s <= s2;
		when s2 =>
			result <= (others => '0');
			rdy <= '0';
			res_mantissa <= (others => '0');
			
			-- Calculate the difference between the expononents
			result_exponent <= big_exponent - small_exponent;
--			a_adder(EXPONENT_LENGTH-1 downto 0) <= big_exponent;
--			b_adder(EXPONENT_LENGTH-1 downto 0) <= small_exponent;
--			ce_adder <= '1';
--			add_adder <= '0';
			
			next_s <= s3;
		when s3 =>
			res_mantissa <= (others => '0');
			result <= (others => '0');
			rdy <= '0';
			normalization <= '0';
			
			-- Shift the smallest mantissa x places to the right if the difference between the exponents are bigger then 0
			-- If the mantissa is shifted, it is denormalized, going from 1.xxxx to 0.xxxx
			if(result_exponent > 0 and result_exponent <= MANTISSA_LENGTH-1) then
				small_mantissa <= std_logic_vector(unsigned(small_mantissa) srl to_integer(unsigned(result_exponent)));
				small_mantissa(MANTISSA_LENGTH-to_integer(unsigned(result_exponent))) <= '1';
				normalization <= '1';
				next_s <= s4;
			elsif(result_exponent > (MANTISSA_LENGTH-1)) then
				-- a >> b
				result_exponent <= big_exponent;
				res_mantissa(MANTISSA_LENGTH-1 downto 0) <= big_mantissa;
				next_s <= s7;
			else
				next_s <= s4;
			end if;
			
		when s4 =>
			result <= (others => '0');
			rdy <= '0';
			
			-- Check for addition or subtraction
			if (operation(0) = '0' or sub_neg = '1') then
				if(normalization = '1') then
					res_mantissa <= ("01" & big_mantissa) + ("00" & small_mantissa);
				else
					res_mantissa <= ("01" & big_mantissa) + ("01" & small_mantissa);
				end if;
				
			elsif (operation(0) = '1') then
				if(normalization = '1') then
					res_mantissa <= ("01" & big_mantissa(MANTISSA_LENGTH-1 downto 0)) - ("00" & small_mantissa(MANTISSA_LENGTH-1 downto 0));
				else
					res_mantissa(MANTISSA_LENGTH-1 downto 0) <= a(MANTISSA_LENGTH-1 downto 0) - b(MANTISSA_LENGTH-1 downto 0);
				end if;
			else
			end if;
			
--			if(normalization = '0') then
--				res_mantissa(MANTISSA_LENGTH) <= '1';
--			else
--				normalization <= '0';
--			end if;
			
			next_s <= s5;
		when s5 =>
			rdy <= '0';
			--res_mantissa <= (others => '0'); 
			
			-- Test if the resulting mantissa is normalized
			if(res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) > 1) then
				-- Needs to be normalized, mantissa bigger then 2
				normalization <= '0';
				
				-- Exponent has to be added
				result_exponent <= big_exponent + res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) - 1;
				-- Resulting mantissa is shifted 1 to right, normalized
				res_mantissa(MANTISSA_LENGTH-1 downto 0) <= res_mantissa(MANTISSA_LENGTH downto 1);
				
				
				-- Resulting mantissa is shifted 1 to right, normalized
--				add_adder <= '1';
--				ce_adder <= '1';
--				a_adder(EXPONENT_LENGTH-1 downto 0) <= big_exponent;
--				if(result_exponent(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) = "10") then
--					b_adder(1 downto 0) <= "01";
--				else
--					b_adder(1 downto 0) <= "10";
--				end if;
				next_s <= s7;
			elsif (res_mantissa(MANTISSA_LENGTH+1 downto MANTISSA_LENGTH) = 1) then
				-- Mantissa is normalized
				normalization <= '0';
				result_exponent <= big_exponent;
				--res_mantissa <= res_mantissa(MANTISSA_LENGTH-1 downto 0);
				next_s <= s7;
				
			else 
				normalization <= '0';
				
				-- Find the position of the leading one
				-- Is this the best way to do it????
				for i in MANTISSA_LENGTH-1 downto 0 loop
					if(res_mantissa(i) = '1') then
					-- Leading one found
						if((MANTISSA_LENGTH-i) > result_exponent) then
							res_mantissa <= (others => '0');
							result_exponent <= (others => '0');
						else
							res_mantissa <= std_logic_vector(unsigned(res_mantissa) sll (MANTISSA_LENGTH-i));
							result_exponent <= big_exponent - (MANTISSA_LENGTH-i);
						end if;
						
--						a_adder(EXPONENT_LENGTH-1 downto 0) <= big_exponent;
--						b_adder(4 downto 0) <= std_logic_vector(to_unsigned(MANTISSA_LENGTH-i,5));
--						add_adder <= '0';
--						ce_adder <= '1';
						exit;
					else
						
					end if;
				end loop;				
				next_s <= s7;
			end if;
			
		-- This state is only needed when exponent has to be changed	
--		when s6 =>
--			a_adder <= (others => '0');
--			b_adder <= (others => '0');
--			result <= (others => '0');
--			ce_adder <= '0';
--			add_adder <= '1';
--			rdy <= '0';
--			
--			big_exponent <= result_exponent(EXPONENT_LENGTH-1 downto 0);
--			next_s <= s7;
			
		when s7 =>
			rdy <= '1';
			
			result(OPERAND_LENGTH-2 downto OPERAND_LENGTH-1-EXPONENT_LENGTH) <= result_exponent;
			result(MANTISSA_LENGTH-1 downto 0) <= res_mantissa(MANTISSA_LENGTH-1 downto 0);
			result(OPERAND_LENGTH-1) <= sign;
			
			next_s <= s0;
	
		when others =>
			result <= (others => '0');
			rdy <= '0';
			res_mantissa <= (others => '0');
			big_exponent <= (others => '0');
			small_exponent <= (others => '0');
			big_mantissa <= (others => '0');
			small_mantissa <= (others => '0');
			
			next_s <= s0;
	end case;
end process;


end Behavioral;

