
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

library work;
use work.all;

entity system is
	generic(
		OPERAND_LENGTH 	: integer := 32;
		EXPONENT_LENGTH 	: integer := 8;
		MANTISSA_LENGTH 	: integer := 23;
		
		EXCEPTIONS			: boolean := false
	);
	port(
		a, b				: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
		operation		:	in std_logic_vector(2 downto 0);
		operation_nd	:	in	std_logic;
		--operation_rfd	:	out std_logic;
		clk				:	in std_logic;
		--clk_a				:	in std_logic;
		reset				:	in std_logic;
		result_ip		: 	out std_logic_vector(OPERAND_LENGTH-1 downto 0);
		--result_conf		: 	out std_logic_vector(OPERAND_LENGTH-1 downto 0);
		rdy				:	out std_logic;
		
		underflow		:	out std_logic;
		overflow			:	out std_logic;
		invalid_op		:	out std_logic;
		divide_by_zero	:	out std_logic
		);
end system;

architecture Behavioral of system is

	signal ready						:	std_logic := '0';
	signal ready_conf_adder_sub	:	std_logic := '0';
	signal ready_conf_mult			:	std_logic := '0';
	signal ready_ip_adder_sub		:	std_logic := '0';
	signal ready_ip_mult				:	std_logic := '0';
	signal ready_ip_div				:	std_logic := '0';
	
	
	signal result_conf_adder_sub	:	std_logic_vector(OPERAND_LENGTH-1 downto 0) := (others => '0');
	signal result_conf_mult			:	std_logic_vector(OPERAND_LENGTH-1 downto 0) := (others => '0');
	signal result_ip_adder_sub		:	std_logic_vector(OPERAND_LENGTH-1 downto 0) := (others => '0');
	signal result_ip_mult			:	std_logic_vector(OPERAND_LENGTH-1 downto 0) := (others => '0');
	signal result_ip_div				:	std_logic_vector(OPERAND_LENGTH-1 downto 0) := (others => '0');
	
	signal fpu_operation				:	std_logic_vector(2 downto 0) := "000";
	signal adder_new_data			:	std_logic := '0';
	signal mult_new_data				:	std_logic := '0';
	signal div_new_data				:	std_logic := '0';
	
	signal sclr_adder_sub			:	std_logic := '1';
	signal sclr_mult					:	std_logic := '1';
	signal sclr_div					:	std_logic := '1';
	
	signal ce_adder_sub				:	std_logic := '0';
	signal ce_mult						:	std_logic := '0';
	signal ce_div						:	std_logic := '0';
	
	-- Comment this if using configurable design
	signal result_conf				: 	std_logic_vector(OPERAND_LENGTH-1 downto 0);
	
	-- Used if EXCEPTIONS is true
	signal underflow_add, underflow_mult, underflow_div				:	std_logic := '0';
	signal overflow_add,	overflow_mult, overflow_div					:	std_logic := '0';
	signal invalid_op_add, invalid_op_mult, invalid_op_div			:	std_logic := '0';
	signal divide_by_zero_div													:	std_logic := '0';

	component fpu_adder_sub
		port(
			a					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			b					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			operation		:	in std_logic_vector(5 downto 0); -- 0 for addition and 1 for subtraction
			operation_nd	:	in	std_logic; -- New Data. Must be set high to indicate that operand A, B and operation is valid
			--clk				:	in std_logic;
			--sclr				:	in std_logic; -- Synchronous Reset. Resets RDY and OPERATION_RFD output. Takes priority over CE
			--ce					:	in std_logic; -- Clock enable
			result			: 	out std_logic_vector(OPERAND_LENGTH-1 downto 0);
			rdy				:	out std_logic -- Set high when result is valid
			--underflow		:	out std_logic;
			--overflow			:	out std_logic;
			--invalid_op		:	out std_logic
		);
	end component;
	
	component fpu_multiplier
		port(
			a					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			b					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			operation_nd	:	in	std_logic;
			--clk				:	in std_logic;
			--sclr				:	in std_logic;
			--ce					:	in std_logic;
			result			: 	out std_logic_vector(OPERAND_LENGTH-1 downto 0);
			rdy				:	out std_logic
			--underflow		:	out std_logic;
			--overflow			:	out std_logic;
			--invalid_op		:	out std_logic
		);
	end component;
	
	component fpu_divider
		port(
			a					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			b					: 	in std_logic_vector(OPERAND_LENGTH-1 downto 0);
			operation_nd	:	in	std_logic;
			--clk				:	in std_logic;
			--sclr				:	in std_logic;
			--ce					:	in std_logic;
			result			: 	out std_logic_vector(OPERAND_LENGTH-1 downto 0);
			rdy				:	out std_logic
			--underflow		:	out std_logic;
			--overflow			:	out std_logic;
			--invalid_op		:	out std_logic;
			--divide_by_zero	:	out std_logic
		);
	end component;
	

begin

		ready <= ready_ip_adder_sub or ready_ip_mult or ready_ip_div;
		rdy <= ready;
		-- ready_conf_adder_sub or 
		
		fpu_operation <= operation(2 downto 0);
		
		underflow <= underflow_add or underflow_mult or underflow_div;
		overflow <= overflow_add or overflow_mult or overflow_div;
		invalid_op <= invalid_op_add or invalid_op_mult or invalid_op_div;
		divide_by_zero <= divide_by_zero_div;
		
		-- Control ready signal
		process(clk, reset, ready_conf_adder_sub, ready_ip_adder_sub, ready_ip_mult, ready_ip_div)
		begin
			if(reset = '1') then
				--sclr_adder_sub <= '0';
--				sclr_mult	 	<= '1';
--				sclr_div	 		<= '1';
			else
				if(rising_edge(clk)) then
					if(ready_conf_adder_sub = '1') then
						result_conf <= result_conf_adder_sub;
					elsif(ready_conf_mult = '1') then
						result_conf <= result_conf_mult;
					elsif(ready_ip_adder_sub = '1') then 
						result_ip <= result_ip_adder_sub;
						--sclr_adder_sub <= '1';
					elsif(ready_ip_mult = '1') then
						result_ip <= result_ip_mult;
					elsif(ready_ip_div = '1') then
						result_ip <= result_ip_div;
					else
					end if;
					
				end if;
			end if;
		end process;
		
		process(clk, reset, fpu_operation, operation_nd, ready)
		begin
			if(reset = '1') then
				sclr_adder_sub <= '1';
				sclr_mult <= '1';
				sclr_div <= '1';
				
				ce_adder_sub <= '0';
				ce_mult <= '0';
				ce_div <= '0';
			else
				if(rising_edge(clk)) then
					if(operation_nd = '1') then
						-- Addition and subtraction
						if(fpu_operation = "000" or fpu_operation ="001") then
							adder_new_data <= '1';
							mult_new_data <= '0';
							div_new_data <= '0';
							
							sclr_adder_sub <= '0';
							sclr_mult <= '1';
							sclr_div <= '1';
							
							ce_adder_sub <= '1';
							ce_mult <= '0';
							ce_div <= '0';
						-- Multiplication
						elsif(fpu_operation = "010") then
							mult_new_data <= '1';
							adder_new_data <= '0';
							div_new_data <= '0';
							
							sclr_adder_sub <= '1';
							sclr_mult <= '0';
							sclr_div <= '1';
							
							ce_adder_sub <= '0';
							ce_mult <= '1';
							ce_div <= '0';
						-- Division
						elsif(fpu_operation = "011") then
							div_new_data <= '1';
							adder_new_data <= '0';
							mult_new_data <= '0';
							
							sclr_adder_sub <= '1';
							sclr_mult <= '1';
							sclr_div <= '0';
							
							ce_adder_sub <= '0';
							ce_mult <= '0';
							ce_div <= '1';
						else
							div_new_data <= '0';
							adder_new_data <= '0';
							mult_new_data <= '0';
							
							sclr_adder_sub <= '1';
							sclr_mult <= '1';
							sclr_div <= '1';
							
							ce_adder_sub <= '0';
							ce_mult <= '0';
							ce_div <= '0';
						end if;
					elsif(ready = '1') then
						sclr_adder_sub <= '1';
						sclr_mult <= '1';
						sclr_div <= '1';
						ce_adder_sub <= '0';
						ce_mult <= '0';
						ce_div <= '0';
						
						adder_new_data <= '0';
						mult_new_data <= '0';
						div_new_data <= '0';
					else
						
					end if;
				end if;
			end if;
		end process;
		
		local_fpu_adder_sub : fpu_adder_sub 
		port map (
			a									=> a, 				-- input [31 : 0] a
			b									=>	b, 				-- input [31 : 0] b
			operation(2 downto 0)		=>	operation, 		-- input [5 : 0] operation
			operation(5 downto 3)		=> "000",
			operation_nd					=>	adder_new_data, 	-- input operation_nd
			--clk								=>	clk, 				-- input clk
			--sclr								=> sclr_adder_sub,
			--ce									=> ce_adder_sub,
			result							=>	result_ip_adder_sub, 		-- output [31 : 0] result
			rdy								=>	ready_ip_adder_sub			-- output rdy	
			--underflow						=> underflow_add,
			--overflow							=> overflow_add,
			--invalid_op						=> invalid_op_add
		);
		
		
		local_fpu_multiplier	:	fpu_multiplier
		port map(
			a					=> a, 				-- input [31 : 0] a
			b					=>	b, 				-- input [31 : 0] b
			operation_nd	=>	mult_new_data, 	-- input operation_nd
			--clk				=>	clk, 				-- input clk
			--sclr				=> sclr_mult,
			--ce					=> ce_mult,
			result			=>	result_ip_mult, 		-- output [31 : 0] result
			rdy				=>	ready_ip_mult			-- output rdy
			--underflow						=> underflow_mult,
			--overflow							=> overflow_mult,
			--invalid_op						=> invalid_op_mult
		);

		local_fpu_divider	:	fpu_divider
		port map(
			a					=> a, 				-- input [31 : 0] a
			b					=>	b, 				-- input [31 : 0] b
			operation_nd	=>	div_new_data, 	-- input operation_nd
			--clk				=>	clk, 				-- input clk
			--sclr				=> sclr_div,
			--ce					=> ce_div,
			result			=>	result_ip_div, 		-- output [31 : 0] result
			rdy				=>	ready_ip_div			-- output rdy
			--underflow						=> underflow_div,
			--overflow							=> overflow_div,
			--invalid_op						=> invalid_op_div,
			--divide_by_zero					=> divide_by_zero_div
		);


end Behavioral;

